
/*
 ******************************************************************
 * HISTORY
 * 15-Oct-94  Jeff Shufelt (js), Carnegie Mellon University
 *      Prepared for 15-681, Fall 1994.
 *
 *
 * altered to implement 4-layer networks by ashwin beeharee 1998
 *
 * altered to allow both 3-layer and 4-layer networks by Steph (sjd) 2001
 *
 ******************************************************************
 */

#ifndef _BACKPROP_H_

#define _BACKPROP_H_

#define BIGRND 0x7fffffff

/*** The neural network data structure.  The network is assumed to
     be a fully-connected feedforward three-layer network.
     Unit 0 in each layer of units is the threshold unit; this means
     that the remaining units are indexed from 1 to n, inclusive.
 ***/

typedef struct {
  int input_n;                  /* number of input units */
  int hidden_n;                 /* number of hidden units */
  int output_n;                 /* number of output units */

  float *input_units;          /* the input units */
  float *hidden_units;         /* the hidden units */
  float *output_units;         /* the output units */

  float *hidden_delta;         /* storage for hidden unit error */
  float *output_delta;         /* storage for output unit error */

  float *target;               /* storage for target vector */

  float **input_weights;       /* weights from input to hidden layer */
  float **hidden_weights;      /* weights from hidden layer to hidden_2 layer */

                                /*** The next two are for momentum ***/
  float **input_prev_weights;  /* previous change on input to hidden wgt */
  float **hidden_prev_weights; /* previous change on hidden to hidden_2 wgt */


  /* additional parameters for second hidden layer added to original code by Steph (sjd) */

  int hidden2_n;
  float *hidden2_units;
  float *hidden2_delta;
  float **hidden2_weights;
  float **hidden2_prev_weights;

} BPNN;


/*** User-level functions ***/

void sjd_bpnn_initialize();

BPNN *sjd_bpnn_create();
void sjd_bpnn_free();

void sjd_bpnn_train();
void sjd_bpnn_feedforward();

float *sjd_bpnn_predict();

void sjd_bpnn_save();
BPNN *sjd_bpnn_read();


#endif

























