// created: 1.8.01 modified: 14.8.01 final clean-up: 24.4.02
// Stephan Dale
// neural network creation and testing program

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "sjd_backprop.h"
#include "nettools.h"

/**************************************************************************************************/

/********************************/
/* external function prototypes */
/********************************/

// from nettools
extern entry *load_netinput(BPNN *net, char *filename);
extern entry *load_single_netinput(BPNN *net, float *data);
extern void display_netindata(entry *head, BPNN *net);
extern void set_netintargetdata(entry *pattern , BPNN *net);
extern void set_netindata(entry *pattern, BPNN *net);
extern void sensitizeinput(char *filename, int method, int chunk);

/********************************/
/* internal function prototypes */
/********************************/
void viewPattern(entry *pattern, BPNN *net);
float  trainIt(entry *pattern, BPNN *net);
float  testIt(entry *pattern, BPNN *net);

/**************************************************************************************************/

void viewPattern(entry *pattern, BPNN *net)
{
  int i, j;

  j = net->input_n;
  printf("input pattern = ");
  for(i=1;i<=j;i++) // the data starts from array element 1
    printf("%f ",pattern->input_vector[i]);
  printf("\n");

  j = net->output_n;
  printf("target pattern = ");
  for(i=1;i<=j;i++)
    printf("%f ",pattern->target_vector[i]);
  printf("\n");

}

// calculate Sum-Squared Error
float calcSSE(float *target, float *prediction, int num_out)
 {
 float diff;
 float error;
 int i;

 error = 0;
 for (i=1; i<=num_out; i++)
   {
   diff = target[i] - prediction[i];
   error += (float)(diff*diff);
   }

 return(error/2.0);
 }


float  trainIt(pattern, net)
entry *pattern;
BPNN *net;
{
  float *prediction;
  float out_err,hid1_err, hid2_err;
  float sum=0;
  float sse;
  /*Setting network Data */
  set_netintargetdata(pattern,net);
  /** Run backprop, learning rate 0.3, momentum 0.3 **/
  /*Training ...... */
  sjd_bpnn_train(net, 0.3, 0.3, &out_err, &hid1_err, &hid2_err);
  prediction = sjd_bpnn_predict(net);
//  return((float)out_err);
  sse = calcSSE(pattern->target_vector, prediction, net->output_n);
  free(prediction);
  return(sse);
}

float  testIt(pattern,net)
entry *pattern;
BPNN *net;
{
  float *prediction;
  float out_err,hid1_err, hid2_err;
  float sum=0;
  float sse;
  /*Setting network Data */
  set_netintargetdata(pattern,net);
//  sjd_bpnn_test(net, 0.3, 0.3, &out_err, &hid1_err, &hid2_err);
//  return((float)out_err);
  prediction = sjd_bpnn_predict(net);
  sse = calcSSE(pattern->target_vector, prediction, net->output_n);
  free(prediction);
  return(sse);
}

/**************************************************************************************************/

int main(int argc, char *argv[])
 {
 int verbose = 0;
 char command;
 char choice;
 char filename[200];
 int i, e;

 // the NN
 BPNN *net = NULL;

 // data streams
 FILE *trainingData, *validationData, *testingData, *testingData1;

 // network variables
 int numInputs;
 int numHidden1;
 int numHidden2;
 int numOutputs;

 // training/testing variables
 int numPatterns, epochs, count;
 float patternError, epochError, sumError;
 entry *headTrainingData, *headValidationData, *headTestingData, *pattern;
 float *output_vector;

 while (command != 'q')
  {
  printf("please enter command (h for help): ");
  scanf("%s", &command);

  switch (command)
    {
    case 'q': // quit
      exit(0);
      break;
    case 'h': // display help
      printf("\n--- help ---\n\n");
      printf(" c - create network\n");
      printf(" l - load network\n");
      printf(" s - save network\n");
      printf(" t - train network\n");
      printf(" e - test network\n");
      printf(" p - print network\n");
      printf(" v - verbose\n");
      printf(" h - help\n");
      printf(" q - quit\n");
      printf("\n------------\n\n");
      break;
    case 'c':

      // get network spec.
      printf("enter number of nodes in input layer: ");
      scanf("%d",&numInputs);
      printf("enter number of nodes in hidden layer 1: ");
      scanf("%d",&numHidden1);
      printf("enter number of nodes in hidden layer 2: ");
      scanf("%d",&numHidden2);
      printf("enter number of nodes in output layer: ");
      scanf("%d",&numOutputs);

      printf("creating %dx%dx%dx%d network\n",numInputs,numHidden1,numHidden2,numOutputs);
      net = sjd_bpnn_create(numInputs,numHidden1,numHidden2,numOutputs);

      break;
    case 'l': // load a network

      if (net != NULL) // if there is already a net, free it's memory
        {
        sjd_bpnn_free(net);
        }

      // load network
      printf("please enter network file name: ");
      scanf("%s", &filename);
      net = sjd_bpnn_read(filename);
      if (net == NULL)
        {
        printf("could not open '%s'\n", filename);
        break;
        }

      break;
    case 's': // save network

      if (net != NULL)
        {
        printf("please enter file name: ");
        scanf("%s", &filename);
        sjd_bpnn_save(net, filename);
        }
      else printf("must create/load a network before saving it\n");

      break;
    case 't': // train network (RMS)

      // check that a network is loaded
      if (net == NULL)
        {
        printf("must create/load a network before training\n");
        break;
        }

      // get desired filenames
      printf("please enter file name of training data: ");
      scanf("%s", &filename);
      headTrainingData = load_netinput(net,filename); // get first pattern of training data (requires net to determine number of inputs)
      if (headTrainingData == NULL)
        {
        printf("could not open '%s'\n", filename);
        break;
        }
      else printf("reading training data from '%s'\n",filename);

      printf("please enter file name of validation data: ");
      scanf("%s", &filename);
      headValidationData = load_netinput(net,filename); // get first pattern of validation data (requires net to determine number of inputs)
      if (headValidationData == NULL)
        {
        printf("could not open '%s'\n", filename);
        break;
        }
      else printf("reading validation data from '%s'\n",filename);

      // get filename to write training error
      printf("please enter name of file to write training error (RMS): ");
      scanf("%s", &filename);
      if ((trainingData = fopen(filename,"w")) == NULL)
        {
        printf("could not open '%s' for writing\n", filename);
        break;
        }
      else printf("writing training error to '%s'\n",filename);

      printf("please enter name of file to write validation error (RMS): ");
      scanf("%s", &filename);
      if ((validationData = fopen(filename,"w")) == NULL)
        {
        printf("could not open '%s' for writing\n", filename);
        break;
        }
      else printf("writing validation error to '%s'\n",filename);

      // get number of epochs
      printf("enter number of epochs : ");
      scanf("%d",&epochs);

      // set total error to 0
      sumError = 0;

      // train
      if (verbose) printf("training epoch ");
      for (e=1; e<=epochs; e++)
	{

        numPatterns = 0;
        if (verbose) printf("%d ",e);
        pattern = headTrainingData; // set first pattern
        epochError = 0; // sum of error for this epoch

        // train for all patterns in training data
        while(pattern/*->next*/ != NULL)
	  {
	  patternError = trainIt(pattern, net); // train network on pattern
	  epochError += patternError; // gather error
	  numPatterns++;
	  pattern = pattern->next;
	  }
	
	// write RMS error for this epoch
	epochError = sqrt(epochError/numPatterns);
	fprintf(trainingData, "%d %f\n", e, epochError);
	
        // test the network for what it has learnt till now
        if (e%10 == 0)
	  {
	  numPatterns = 0;
	  if (verbose) printf("validating ",e);
	  pattern = headValidationData; // set first validation pattern
	  epochError = 0;
	
	  while (pattern/*->next*/!=NULL)
	    {
	    numPatterns++;
	    patternError = testIt(pattern, net);
            epochError += patternError;
 	    pattern = pattern->next;
	    }
	
          epochError = sqrt(epochError/numPatterns);
          fprintf(validationData, "%d %f\n", e, epochError); // write RMS error
          }

        } // end of training
      if (verbose) printf("\n");

      fclose(trainingData);
      fclose(validationData);
      break;
    case 'e': // test a network (get SSE and output)

      // check that a network is loaded
      if (net == NULL)
        {
        printf("must create/load a network before testing\n");
        break;
        }

      // get desired testing data filename
      printf("please enter file name of testing data: ");
      scanf("%s", &filename);
      headTestingData = load_netinput(net,filename); // get first pattern of training data (requires net to determine number of inputs)
      if (headTestingData == NULL)
        {
        printf("could not open '%s'\n", filename);
        break;
        }
      else printf("reading testing data from '%s'\n",filename);

      // get filename to write network output
      printf("please enter name of file to write targets and network outputs: ");
      scanf("%s", &filename);
      if ((testingData = fopen(filename,"w")) == NULL)
        {
        printf("could not open '%s' for writing\n", filename);
        break;
        }
      else printf("writing targets and network outputs to '%s'\n",filename);

      // get filename to write testing error
      printf("please enter name of file to write testing error (SSE): ");
      scanf("%s", &filename);
      if ((testingData1 = fopen(filename,"w")) == NULL)
        {
        printf("could not open '%s' for writing\n", filename);
        fclose(testingData);
        break;
        }
      else printf("writing testing error to '%s'\n",filename);

      pattern = headTestingData;
      count = 0;
      while (pattern/*->next*/!=NULL)
	{
        count++;
	patternError = testIt(pattern, net); // get SSE
        output_vector = sjd_bpnn_predict(net);
        if (verbose)
          {
          printf("pattern %d; input =", count);
          for (i=1; i<=net->input_n; i++)
            printf(" %f", pattern->input_vector[i]);
          printf(", desired output =");
          for (i=1; i<=net->output_n; i++)
            printf(" %f", pattern->target_vector[i]);
          printf(", network output =");
          for (i=1; i<=net->output_n; i++)
            printf(" %f", output_vector[i]);
          printf(", error = %f\n", patternError);
          }
        fprintf(testingData, "%d ", count);
        for (i=1; i<=net->output_n; i++)
          fprintf(testingData, "%f ", pattern->target_vector[i]);
        for (i=1; i<=net->output_n; i++)
          fprintf(testingData, "%f ", output_vector[i]);
        fprintf(testingData, "\n");
        fprintf(testingData1, "%d %f\n", count, patternError); // write SSE
        pattern = pattern->next;
	}

      fclose(testingData);	
      fclose(testingData1);	

      break;
    case 'p':
      if (net != NULL) sjd_print_network(net);
      else printf("must create/load network first\n");
      break;
    case 'v':
      verbose = !verbose;
      if (verbose)
        printf("verbose on\n");
      else
        printf("verbose off\n");
      break;
    }
  }

 return 0;
 }

/**************************************************************************************************/

