
/**************************************************************************************************
 * HISTORY
 * 15-Oct-94  Jeff Shufelt (js), Carnegie Mellon University
 *	Prepared for 15-681, Fall 1994.
 *
 *
 * altered to implement 4-layer networks by ashwin beeharee 1998
 *
 * altered to allow both 3-layer and 4-layer networks by Steph (sjd) 2001
 * (by adding conditions to test whether any nodes in 2nd hidden layer)
 *
 **************************************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include "sjd_backprop.h"
#include <math.h>


#define ABS(x)          (((x) > 0.0) ? (x) : (-(x)))

#define fastcopy(to,from,len)\
{\
  register char *_to,*_from;\
  register int _i,_l;\
  _to = (char *)(to);\
  _from = (char *)(from);\
  _l = (len);\
  for (_i = 0; _i < _l; _i++) *_to++ = *_from++;\
}

/**************************************************************************************************/

/********************************/
/* external function prototypes */
/********************************/

// from JS's original backprop.c
extern float **alloc_2d_flt();
extern float drnd();
extern float dpn1();
extern float squash();
extern float *alloc_1d_flt();
extern void bpnn_randomize_weights();
extern void bpnn_zero_weights();
extern void bpnn_initialize();
extern void bpnn_layerforward();
extern void bpnn_output_error();
extern void bpnn_hidden_error();
extern void bpnn_adjust_weights();

/**************************************************************************************************/

BPNN *sjd_bpnn_internal_create(n_in, n_hidden_1, n_hidden_2, n_out)
int n_in, n_hidden_1, n_hidden_2, n_out;
{
  BPNN *newnet;

  newnet = (BPNN *) malloc (sizeof (BPNN));
  if (newnet == NULL) {
    printf("BPNN_CREATE: Couldn't allocate neural network\n");
    return (NULL);
  }

  newnet->input_n = n_in;
  newnet->hidden_n = n_hidden_1;
  newnet->hidden2_n = n_hidden_2;
  newnet->output_n = n_out;

  newnet->input_units = alloc_1d_flt(n_in + 1);
  newnet->hidden_units = alloc_1d_flt(n_hidden_1 + 1);
  newnet->hidden2_units = alloc_1d_flt(n_hidden_2 + 1);
  newnet->output_units = alloc_1d_flt(n_out + 1);

  newnet->hidden_delta = alloc_1d_flt(n_hidden_1 + 1);
  newnet->hidden2_delta = alloc_1d_flt(n_hidden_2 + 1);
  newnet->output_delta = alloc_1d_flt(n_out + 1);
  newnet->target = alloc_1d_flt(n_out + 1);

  newnet->input_weights = alloc_2d_flt(n_in + 1, n_hidden_1 + 1);
  newnet->input_prev_weights = alloc_2d_flt(n_in + 1, n_hidden_1 + 1);

  if (n_hidden_2==0)
    {
    newnet->hidden_weights = alloc_2d_flt(n_hidden_1 + 1, n_out + 1);
    newnet->hidden_prev_weights = alloc_2d_flt(n_hidden_1 + 1, n_out + 1);
    }
  else
    {
    newnet->hidden_weights = alloc_2d_flt(n_hidden_1 + 1, n_hidden_2 + 1);
    newnet->hidden2_weights = alloc_2d_flt(n_hidden_2 + 1, n_out + 1);
    newnet->hidden_prev_weights = alloc_2d_flt(n_hidden_1 + 1, n_hidden_2 + 1);
    newnet->hidden2_prev_weights = alloc_2d_flt(n_hidden_2 + 1, n_out + 1);
    }

  return (newnet);
}


void sjd_bpnn_free(net)
BPNN *net;
{
  int n1, n2, n3, i;

  n1 = net->input_n;
  n2 = net->hidden_n;
  n3 = net->hidden2_n;

  free((char *) net->input_units);
  free((char *) net->hidden_units);
  free((char *) net->hidden2_units);
  free((char *) net->output_units);

  free((char *) net->hidden_delta);
  free((char *) net->hidden2_delta);
  free((char *) net->output_delta);
  free((char *) net->target);

  for (i = 0; i <= n1; i++) {
    free((char *) net->input_weights[i]);
    free((char *) net->input_prev_weights[i]);
  }
  free((char *) net->input_weights);
  free((char *) net->input_prev_weights);

  for (i = 0; i <= n2; i++) {
    free((char *) net->hidden_weights[i]);
    free((char *) net->hidden_prev_weights[i]);
  }

  free((char *) net->hidden_weights);
  free((char *) net->hidden_prev_weights);

  if (n3 != 0) // free hidden weights if any were allocated in the first place
   {
    for (i = 0; i <= n3; i++) {
      free((char *) net->hidden2_weights[i]);
      free((char *) net->hidden2_prev_weights[i]);
    }

    free((char *) net->hidden2_weights);
    free((char *) net->hidden2_prev_weights);
   }

  free((char *) net);
}


/*** Creates a new fully-connected network from scratch,
     with the given numbers of input, hidden, and output units.
     Threshold units are automatically included.  All weights are
     randomly initialized.

     Space is also allocated for temporary storage (momentum weights,
     error computations, etc).
***/

BPNN *sjd_bpnn_create(n_in,n_hidden_1,n_hidden_2,n_out)
int n_in, n_hidden_1, n_hidden_2, n_out;
{

  BPNN *newnet;

//  printf("sjd_bpnn_create %d %d %d %d\n",n_in,n_hidden_1,n_hidden_2,n_out);
  newnet = sjd_bpnn_internal_create(n_in,n_hidden_1,n_hidden_2,n_out);

#ifdef INITZERO
  bpnn_zero_weights(newnet->input_weights, n_in, n_hidden_1);
#else
  bpnn_randomize_weights(newnet->input_weights, n_in, n_hidden_1);
#endif
  bpnn_zero_weights(newnet->input_prev_weights, n_in, n_hidden_1);

  if (n_hidden_2==0) // if there is no second hidden layer
    {
    bpnn_randomize_weights(newnet->hidden_weights, n_hidden_1, n_out);
    bpnn_zero_weights(newnet->hidden_prev_weights, n_hidden_1, n_out);
    }
  else
    {
    bpnn_randomize_weights(newnet->hidden_weights, n_hidden_1, n_hidden_2);
    bpnn_randomize_weights(newnet->hidden2_weights, n_hidden_2, n_out);
    bpnn_zero_weights(newnet->hidden_prev_weights, n_hidden_1, n_hidden_2);
    bpnn_zero_weights(newnet->hidden2_prev_weights, n_hidden_2, n_out);
    }
  return (newnet);
}

void sjd_bpnn_feedforward(net)
BPNN *net;
{
  int in, hid, hid2, out;

  in = net->input_n;
  hid = net->hidden_n;
  hid2 = net->hidden2_n;
  out = net->output_n;

  /*** Feed forward input activations. ***/
  bpnn_layerforward(net->input_units, net->hidden_units,
      net->input_weights, in, hid);
  if (hid2==0)
    {
    bpnn_layerforward(net->hidden_units, net->output_units,
      net->hidden_weights, hid, out);
    }
  else
    {
    bpnn_layerforward(net->hidden_units, net->hidden2_units,
      net->hidden_weights, hid, hid2);
    bpnn_layerforward(net->hidden2_units, net->output_units,
      net->hidden2_weights, hid2, out);
    }

}


void sjd_bpnn_train(net, eta, momentum, eo, eh, eh2)
BPNN *net;
float eta, momentum, *eo, *eh, *eh2;
{
  int i;

  int in, hid, hid2, out;
  float out_err, hid_err, hid2_err;

  in = net->input_n;
  hid = net->hidden_n;
  hid2 = net->hidden2_n;
  out = net->output_n;

  /*** Feed forward input activations. ***/
  bpnn_layerforward(net->input_units, net->hidden_units,
      net->input_weights, in, hid);
  if (hid2==0)
    {
    bpnn_layerforward(net->hidden_units, net->output_units,
      net->hidden_weights, hid, out);
    }
  else
    {
    bpnn_layerforward(net->hidden_units, net->hidden2_units,
      net->hidden_weights, hid, hid2);
    bpnn_layerforward(net->hidden2_units, net->output_units,
      net->hidden2_weights, hid2, out);
    }

  /*** Compute error on output and hidden units. ***/
  bpnn_output_error(net->output_delta, net->target, net->output_units,
      out, &out_err);
  if (hid2==0)
    {
    bpnn_hidden_error(net->hidden_delta, hid, net->output_delta, out,
      net->hidden_weights, net->hidden_units, &hid_err);
    }
  else
    {
    bpnn_hidden_error(net->hidden2_delta, hid2, net->output_delta, out,
      net->hidden2_weights, net->hidden2_units, &hid2_err);
    bpnn_hidden_error(net->hidden_delta, hid, net->hidden2_delta, hid2,
      net->hidden_weights, net->hidden_units, &hid_err);
    }
  *eo = out_err;
  *eh2 = hid2_err;
  *eh = hid_err;

  /*** Adjust input and hidden weights. ***/
  if (hid2==0)
    {
    bpnn_adjust_weights(net->output_delta, out, net->hidden_units, hid,
      net->hidden_weights, net->hidden_prev_weights, eta, momentum);
    }
  else
    {
    bpnn_adjust_weights(net->output_delta, out, net->hidden2_units, hid2,
      net->hidden2_weights, net->hidden2_prev_weights, eta, momentum);
    bpnn_adjust_weights(net->hidden2_delta, hid2, net->hidden_units, hid,
      net->hidden_weights, net->hidden_prev_weights, eta, momentum);
    }
  bpnn_adjust_weights(net->hidden_delta, hid, net->input_units, in,
      net->input_weights, net->input_prev_weights, eta, momentum);

}

/*

Same as train except no weight ajustements
*/



void sjd_bpnn_test(net, eta, momentum, eo, eh, eh2)
BPNN *net;
float eta, momentum, *eo, *eh, *eh2;
{
  int in, hid, hid2, out;
  float out_err, hid_err, hid2_err;

  in = net->input_n;
  hid = net->hidden_n;
  hid2 = net->hidden2_n;
  out = net->output_n;

  /*** Feed forward input activations. ***/

  bpnn_layerforward(net->input_units, net->hidden_units,
      net->input_weights, in, hid);
  if (hid2==0)
    {
    bpnn_layerforward(net->hidden_units, net->output_units,
      net->hidden_weights, hid, out);
    }
  else
    {
    bpnn_layerforward(net->hidden_units, net->hidden2_units,
      net->hidden_weights, hid, hid2);
    bpnn_layerforward(net->hidden2_units, net->output_units,
      net->hidden2_weights, hid2, out);
    }


  /*** Compute error on output and hidden units. ***/
  bpnn_output_error(net->output_delta, net->target, net->output_units,
      out, &out_err);
  if (hid2==0)
    {
    bpnn_hidden_error(net->hidden_delta, hid, net->output_delta, out,
      net->hidden_weights, net->hidden_units, &hid_err);
    }
  else
    {
    bpnn_hidden_error(net->hidden2_delta, hid2, net->output_delta, out,
      net->hidden2_weights, net->hidden2_units, &hid2_err);
    bpnn_hidden_error(net->hidden_delta, hid, net->hidden2_delta, hid2,
      net->hidden_weights, net->hidden_units, &hid_err);
    }
  *eo = out_err;
  *eh2 = hid2_err;
  *eh = hid_err;


}



float *sjd_bpnn_predict(net/*, eta, momentum*/)
BPNN *net;
//float eta, momentum;

{
  int in, hid, hid2, out,i;
  float out_err, hid_err, hid2_err;
  float *outdata;



  in = net->input_n;
  hid = net->hidden_n;
  hid2 = net->hidden2_n;
  out = net->output_n;
  outdata=alloc_1d_flt(out + 1);

  /*** Feed forward input activations. ***/
  bpnn_layerforward(net->input_units, net->hidden_units,
      net->input_weights, in, hid);
  if (hid2==0)
    {
    bpnn_layerforward(net->hidden_units, net->output_units,
      net->hidden_weights, hid, out);
    }
  else
    {
    bpnn_layerforward(net->hidden_units, net->hidden2_units,
      net->hidden_weights, hid, hid2);
    bpnn_layerforward(net->hidden2_units, net->output_units,
      net->hidden2_weights, hid2, out);
    }

  for(i=1;i<=out;i++)
    outdata[i]=net->output_units[i];
  return(outdata);

}


void sjd_bpnn_save(net, filename)
BPNN *net;
char *filename;
{
  int fd, n1, n2, n3, n4, i, j, memcnt;
  float dvalue, **w;
  char *mem;

  if ((fd = creat(filename, 0644)) == -1) {
    printf("BPNN_SAVE: Cannot create '%s'\n", filename);
    return;
  }

  n1 = net->input_n; n2 = net->hidden_n; n3 = net->hidden2_n; n4 = net->output_n;
  printf("saving %dx%dx%dx%d network to '%s'\n", n1, n2, n3, n4, filename);
  fflush(stdout);

  write(fd, (char *) &n1, sizeof(int));
  write(fd, (char *) &n2, sizeof(int));
  write(fd, (char *) &n3, sizeof(int));
  write(fd, (char *) &n4, sizeof(int));

  memcnt = 0;
  w = net->input_weights;
  mem = (char *) malloc ((unsigned) ((n1+1) * (n2+1) * sizeof(float)));
  for (i = 0; i <= n1; i++) {
    for (j = 0; j <= n2; j++) {
      dvalue = w[i][j];
      fastcopy(&mem[memcnt], &dvalue, sizeof(float));
      memcnt += sizeof(float);
    }
  }
  write(fd, mem, (n1+1) * (n2+1) * sizeof(float));
  free(mem);

  if (n3==0)
    {
    memcnt = 0;
    w = net->hidden_weights;
    mem = (char *) malloc ((unsigned) ((n2+1) * (n4+1) * sizeof(float)));
    for (i = 0; i <= n2; i++)
      {
      for (j = 0; j <= n4; j++)
        {
        dvalue = w[i][j];
        fastcopy(&mem[memcnt], &dvalue, sizeof(float));
        memcnt += sizeof(float);
        }
      }
    write(fd, mem, (n2+1) * (n4+1) * sizeof(float));
    free(mem);
    }
  else
    {
    memcnt = 0;
    w = net->hidden_weights;
    mem = (char *) malloc ((unsigned) ((n2+1) * (n3+1) * sizeof(float)));
    for (i = 0; i <= n2; i++)
      {
      for (j = 0; j <= n3; j++)
        {
        dvalue = w[i][j];
        fastcopy(&mem[memcnt], &dvalue, sizeof(float));
        memcnt += sizeof(float);
        }
      }
    write(fd, mem, (n2+1) * (n3+1) * sizeof(float));
    free(mem);

    memcnt = 0;
    w = net->hidden2_weights;
    mem = (char *) malloc ((unsigned) ((n3+1) * (n4+1) * sizeof(float)));
    for (i = 0; i <= n3; i++)
      {
      for (j = 0; j <= n4; j++)
        {
        dvalue = w[i][j];
        fastcopy(&mem[memcnt], &dvalue, sizeof(float));
        memcnt += sizeof(float);
        }
      }
    write(fd, mem, (n3+1) * (n4+1) * sizeof(float));
    free(mem);
    }

/*  print_network(net); */
  close(fd);
  return;
}


BPNN *sjd_bpnn_read(filename)
char *filename;
{
  char *mem;
  BPNN *new;
  int fd, n1, n2, n3, n4, i, j, memcnt;

  if ((fd = open(filename, 0, 0644)) == -1) {
    return (NULL);
  }

  printf("reading '%s'\n", filename);  fflush(stdout);

  read(fd, (char *) &n1, sizeof(int));
  read(fd, (char *) &n2, sizeof(int));
  read(fd, (char *) &n3, sizeof(int));
  read(fd, (char *) &n4, sizeof(int));
  new = sjd_bpnn_internal_create(n1, n2, n3, n4);

  printf("'%s' contains a %dx%dx%dx%d network\n", filename, n1, n2, n3, n4);
  printf("reading input weights...");  fflush(stdout);

  memcnt = 0;
  mem = (char *) malloc ((unsigned) ((n1+1) * (n2+1) * sizeof(float)));
  read(fd, mem, (n1+1) * (n2+1) * sizeof(float));
  for (i = 0; i <= n1; i++) {
    for (j = 0; j <= n2; j++) {
      fastcopy(&(new->input_weights[i][j]), &mem[memcnt], sizeof(float));
      memcnt += sizeof(float);
    }
  }
  free(mem);
  printf("done\nreading hidden weights...");  fflush(stdout);

  if (n3==0)
    {
    memcnt = 0;
    mem = (char *) malloc ((unsigned) ((n2+1) * (n4+1) * sizeof(float)));
    read(fd, mem, (n2+1) * (n4+1) * sizeof(float));
    for (i = 0; i <= n2; i++)
      {
      for (j = 0; j <= n4; j++)
        {
        fastcopy(&(new->hidden_weights[i][j]), &mem[memcnt], sizeof(float));
        memcnt += sizeof(float);
        }
      }
    free(mem);
    printf("done\nreading hidden layer 1 weights...");  fflush(stdout);
    close(fd);
    }
  else
    {
    memcnt = 0;
    mem = (char *) malloc ((unsigned) ((n2+1) * (n3+1) * sizeof(float)));
    read(fd, mem, (n2+1) * (n3+1) * sizeof(float));
    for (i = 0; i <= n2; i++)
      {
      for (j = 0; j <= n3; j++)
        {
        fastcopy(&(new->hidden_weights[i][j]), &mem[memcnt], sizeof(float));
        memcnt += sizeof(float);
        }
      }
    free(mem);
    printf("done\nreading hidden layer 1 weights...");  fflush(stdout);

    memcnt = 0;
    mem = (char *) malloc ((unsigned) ((n3+1) * (n4+1) * sizeof(float)));
    read(fd, mem, (n3+1) * (n4+1) * sizeof(float));
    for (i = 0; i <= n3; i++)
      {
      for (j = 0; j <= n4; j++)
        {
        fastcopy(&(new->hidden2_weights[i][j]), &mem[memcnt], sizeof(float));
        memcnt += sizeof(float);
        }
      }
    free(mem);
    printf("done\nreading hidden layer 2 weights...");  fflush(stdout);
    close(fd);
    }

/*  print_network(new); */
  printf("done\n");  fflush(stdout);

  bpnn_zero_weights(new->input_prev_weights, n1, n2);
  if (n3==0)
    {
    bpnn_zero_weights(new->hidden_prev_weights, n2, n4);
    }
  else
    {
    bpnn_zero_weights(new->hidden_prev_weights, n2, n3);
    bpnn_zero_weights(new->hidden2_prev_weights, n3, n4);
    }

  return (new);
}


void sjd_print_network(net)

BPNN *net;

{
register i,j;
int n1,n2,n3,n4;

  n1 = net->input_n;
  n2 = net->hidden_n;
  n3 = net->hidden2_n;
  n4 = net->output_n;

  printf("network\n");
  printf("input weights\n");
  for (i = 1; i <= n1; i++) {
    for (j = 1; j <= n2; j++)
     printf(" %8.3f",net->input_weights[i][j]);
    printf("\n");
  }

  if (n3==0)
    {
    printf("first hidden weights\n");
    for (i = 1; i <= n2; i++)
      {
      for (j = 1; j <= n4; j++)
        printf(" %8.3f",net->hidden_weights[i][j]);
       printf("\n");
      }
    }
  else
    {
    printf("first hidden weights\n");
    for (i = 1; i <= n2; i++)
      {
      for (j = 1; j <= n3; j++)
        printf(" %8.3f",net->hidden_weights[i][j]);
       printf("\n");
      }

    printf("second hidden weights\n");
    for (i = 1; i <= n3; i++)
      {
      for (j = 1; j <= n4; j++)
        printf(" %8.3f",net->hidden2_weights[i][j]);
      printf("\n");
      }
    }


}







