// created: 5.8.01 modified: 23.8.01 final clean-up: 24.4.02
// Stephan Dale
// avatarlib IO library

#include <stdlib.h>
#include "avatarlibio.h"
#include "avatarlib.h" // needed for reading arm data

/**************************************************************************************************/

/**********************/
/* external variables */
/**********************/

// avatar positional data
extern MAV_vector leftHand, rightHand; // the position of the avatars hands

// tracked data
extern float leftHandRoll, leftHandPitch, leftHandYaw, rightHandRoll, rightHandPitch, rightHandYaw;
extern float leftArmRollTracked, leftArmPitchTracked, leftArmYawTracked, rightArmRollTracked, rightArmPitchTracked, rightArmYawTracked;
// neural network data
extern float leftArmRollNetwork, leftArmPitchNetwork, leftArmYawNetwork, rightArmRollNetwork, rightArmPitchNetwork, rightArmYawNetwork;

/************/
/* globals */
/***********/

char trainingFilename[200];
char testingFilename[200];

FILE *trainingData;
FILE *testingData;

/**************************************************************************************************/

/*******************************/
/* set training data filenames */
/*******************************/
void setTrainingFilenames(void)
 {
 printf("please enter file name to write training data: ");
 scanf("%s", trainingFilename);
 }

/******************************/
/* open training data streams */
/******************************/
int openTrainingFiles(void)
 {
 int error = 1;

 if ((trainingData = fopen(trainingFilename,"w"))!=NULL)
   {
   printf("writing training data to \"%s\"\n", trainingFilename);
   }
 else
   {
   printf("failed to open %s for writing\n", trainingFilename);
   error = 0;
   }
 return error;
 }

/*********************************/
/* write training data to a file */
/*********************************/
void writeTrainingData(void *unused)
 {
 // input is hand position relative to shoulder and hand RPY relative to back rotation
 // desired output is yaw and pitch relative to shoulder
 fprintf(trainingData, "%f %f %f %f %f %f %f %f\n",
 	rightHand.x, rightHand.y, rightHand.z,
 	rightHandRoll, rightHandPitch, rightHandYaw,
	rightArmYawTracked, rightArmPitchTracked);
 }

/*******************************/
/* close training file streams */
/*******************************/
void closeTrainingFile(void)
 {
 printf("closing training data stream\n");
 fclose(trainingData);
 }

/**************************************************************************************************/

/******************************/
/* set testing data filenames */
/******************************/
void setTestingFilenames(void)
 {
 printf("please enter file name to write testing data: ");
 scanf("%s", testingFilename);
 }

/*****************************/
/* open testng data streams */
/*****************************/
int openTestingFiles(void)
 {
 int error = 1;
 if ((testingData = fopen(testingFilename,"w"))!=NULL)
   {
   printf("writing testing data to \"%s\"\n", testingFilename);
   }
 else
   {
   printf("failed to open %s for writing\n", testingFilename);
   error = 0;
   }
 return error;
 }

/****************************************/
/* write testing results data to a file */
/****************************************/
void writeTestingData(void *unused)
 {
 // NN prediction of upper arm rotation
 fprintf(testingData, "%f %f %f %f\n", rightArmYawTracked, rightArmPitchTracked, rightArmYawNetwork, rightArmPitchNetwork);
 }

/******************************/
/* close testing file streams */
/******************************/
void closeTestingFiles(void)
 {
 printf("closing testing data streams\n");
 fclose(testingData);
 }

/**************************************************************************************************/

